@@ -1,4 +1,4 @@
-/* Copyright 1999, 2000 Red Hat, Inc.
+/* Copyright 1999-2003 Red Hat, Inc.
  *
  * This software may be freely redistributed under the terms of the GNU
  * public license.
@@ -18,12 +18,13 @@ struct ddcDevice {
     /* common fields */
     struct device *next;	/* next device in list */
 	int index;
-    enum deviceClass class;	/* type */
+    enum deviceClass type;	/* type */
     enum deviceBus bus;		/* bus it's attached to */
     char * device;		/* device file associated with it */
     char * driver;		/* driver to load, if any */
     char * desc;		/* a description */
     int detached;
+    void * classprivate;
     /* ddc-specific fields */
     struct ddcDevice *(*newDevice) (struct ddcDevice *dev);
     void (*freeDevice) (struct ddcDevice *dev);
