@@ -1,4 +1,4 @@
-/* Copyright 1999, 2000 Red Hat, Inc.
+/* Copyright 1999-2003 Red Hat, Inc.
  *
  * This software may be freely redistributed under the terms of the GNU
  * public license.
@@ -16,6 +16,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <libgen.h>
 
 #include <sys/stat.h>
 #include <sys/utsname.h>
@@ -83,7 +84,8 @@ static void pciFreeDevice(struct pciDevice *dev) {
 
 static void pciWriteDevice(FILE *file, struct pciDevice *dev) {
 	writeDevice(file, (struct device *)dev);
-	fprintf(file,"vendorId: %04x\ndeviceId: %04x\nsubVendorId: %04x\nsubDeviceId: %04x\npciType: %d\n",dev->vendorId,dev->deviceId,dev->subVendorId,dev->subDeviceId,dev->pciType);
+	fprintf(file,"vendorId: %04x\ndeviceId: %04x\nsubVendorId: %04x\nsubDeviceId: %04x\npciType: %d\npcidom: %4x\npcibus: %2x\npcidev: %2x\npcifn: %2x\n",
+		dev->vendorId,dev->deviceId,dev->subVendorId,dev->subDeviceId,dev->pciType,dev->pcidom,dev->pcibus,dev->pcidev,dev->pcifn);
 }
 
 static int pciCompareDevice(struct pciDevice *dev1, struct pciDevice *dev2)
@@ -114,6 +116,10 @@ struct pciDevice * pciNewDevice(struct pciDevice *dev) {
 	ret->pciType = dev->pciType;
 	ret->subVendorId = dev->subVendorId;
         ret->subDeviceId = dev->subDeviceId;
+	ret->pcidom = dev->pcidom;
+	ret->pcibus = dev->pcibus;
+	ret->pcidev = dev->pcidev;
+	ret->pcifn = dev->pcifn;
     } else {
 	ret->pciType = PCI_UNKNOWN;
     }
@@ -157,8 +163,12 @@ static unsigned int kudzuToPci(enum deviceClass class) {
 	return PCI_CLASS_INPUT_MOUSE;
      case CLASS_USB:
 	return PCI_CLASS_SERIAL_USB;
+     case CLASS_FIREWIRE:
+	return PCI_CLASS_SERIAL_FIREWIRE;
      case CLASS_SOCKET:
 	return PCI_CLASS_BRIDGE_CARDBUS;
+     case CLASS_IDE:
+	return PCI_CLASS_STORAGE_IDE;
      default:
 	return 0;
     }
@@ -170,6 +180,8 @@ static enum deviceClass pciToKudzu(unsigned int class) {
     switch (class >> 8) {
      case PCI_BASE_CLASS_DISPLAY:
 	return CLASS_VIDEO;
+    case PCI_BASE_CLASS_NETWORK:
+	return CLASS_NETWORK;
      default:
 	break;
     }
@@ -198,6 +210,8 @@ static enum deviceClass pciToKudzu(unsigned int class) {
 	return CLASS_SCSI;
      case PCI_CLASS_SERIAL_USB:
 	return CLASS_USB;
+     case PCI_CLASS_SERIAL_FIREWIRE:
+	return CLASS_FIREWIRE;
      case PCI_CLASS_BRIDGE_CARDBUS:
 	return CLASS_SOCKET;
      case PCI_CLASS_NETWORK_ETHERNET:
@@ -208,6 +222,8 @@ static enum deviceClass pciToKudzu(unsigned int class) {
 	return CLASS_NETWORK;
      case PCI_CLASS_NETWORK_ATM:
 	return CLASS_NETWORK;
+     case PCI_CLASS_STORAGE_IDE:
+	return CLASS_IDE;
      default:
 	return CLASS_OTHER;
     }
@@ -246,9 +262,11 @@ int pciReadDrivers(char *filename) {
 	char *descrip;
 	char path[256];
 	struct utsname utsbuf;
+	char *dir = NULL;
      
 
 	if (filename) {
+		dir = dirname(strdup(filename));
 		fd = open(filename, O_RDONLY);
 		if (fd < 0)
 		  return -1;
@@ -267,6 +285,8 @@ int pciReadDrivers(char *filename) {
 		}
 	}
 	buf = bufFromFd(fd);
+	if (!buf) goto pcimap;
+	
 
 	/* upper bound */
 	numDrivers = 1;
@@ -364,7 +384,7 @@ int pciReadDrivers(char *filename) {
 			nextDevice->desc = strdup(descrip);
 			nextDevice->next = NULL;
 			nextDevice->device = NULL;
-			nextDevice->class = 0;
+			nextDevice->type = 0;
 			nextDevice->bus = BUS_PCI;
 			if (merge && tmpdev)
 			  nextDevice = tmpdev;
@@ -383,9 +403,16 @@ int pciReadDrivers(char *filename) {
 	free(buf);
 
 	qsort(pciDeviceList, numPciDevices, sizeof(*pciDeviceList), devCmp);
-	
+
+pcimap:
+	fd = -1;
 	/* OK, now read the modutils modules.pcimap file. */
-	if (!filename) {
+	if (filename) {
+		snprintf(path,255,"%s/modules.pcimap",dir);
+		fd = open(path, O_RDONLY);
+		if (fd < 0)
+			return 0;
+	} else {
 		uname(&utsbuf);
 		if (strstr(utsbuf.release,"BOOT")) {
 			char kernelver[64];
@@ -410,156 +437,161 @@ int pciReadDrivers(char *filename) {
 				}
 			}
 		}
-		bufptr = bufFromFd(fd);
+	}
+	bufptr = bufFromFd(fd);
+	if (!bufptr) return 0;
 
-		start = buf = bufptr;
-		numDrivers = 0;
-		while ((start = strchr(start, '\n'))) {
-			numDrivers++;
-			start++;
+	start = buf = bufptr;
+	numDrivers = 0;
+	while ((start = strchr(start, '\n'))) {
+		numDrivers++;
+		start++;
+	}
+	pciDeviceList = realloc(pciDeviceList, sizeof(*pciDeviceList) *
+				(numPciDevices + numDrivers));
+	nextDevice = pciDeviceList + numPciDevices;
+
+	start = buf = bufptr;
+	while (start && *start) {
+		while (*buf && *buf != '\n') buf++;
+		if (*buf) {
+			*buf = '\0';
+			buf++;
 		}
-		pciDeviceList = realloc(pciDeviceList, sizeof(*pciDeviceList) *
-					(numPciDevices + numDrivers));
-	 	nextDevice = pciDeviceList + numPciDevices;
-	
-		start = buf = bufptr;
-		while (start && *start) {
-			while (*buf && *buf != '\n') buf++;
-			if (*buf) {
-				*buf = '\0';
-				buf++;
-			}
-			if (*start == '#') {
-				start = buf;
-				continue;
-			}
-			ptr = start;
-			while (*ptr && *ptr != ' ') ptr++;
-			if (*ptr) {
-				*ptr = '\0';
-				ptr++;
-			}
-			module = strdup(start);
-			while (*ptr && *ptr == ' ') ptr++;
-			if (!*ptr) {
-				start = buf;
-				free(module);
-				continue;
-			}
-			start = ptr;
-			while (*ptr && *ptr != ' ') ptr++;
-			if (*ptr) {
-				*ptr = '\0';
-				ptr++;
-			}
-			vendid = strtoul(start,(char **)NULL,16);
-			if (vendid == 0xffffffff) {
-				start = buf;
-				free(module);
-				continue;
-			}
-			while (*ptr && *ptr == ' ') ptr++;
-			if (!*ptr) {
-				start = buf;
-				free(module);
-				continue;
-			}
-			start = ptr;
-			while (*ptr && *ptr != ' ') ptr++;
-			if (*ptr) {
-				*ptr = '\0';
-				ptr++;
-			}
-			devid = strtoul(start, (char **)NULL, 16);
-			while (*ptr && *ptr == ' ') ptr++;
-			if (!*ptr) {
-				start = buf;
-				free(module);
-				continue;
-			}
-			start = ptr;
-			while (*ptr && *ptr != ' ') ptr++;
-			if (*ptr) {
-				*ptr = '\0';
-				ptr++;
-			}
-			subvendid = strtoul(start, (char **)NULL, 16);
-			if (subvendid == 0xffffffff) {
-				subvendid = 0xffff;
-				subdevid = 0;
-			} else {
-				while (*ptr && *ptr == ' ') ptr++;
-				if (!*ptr) {
-					start = buf;
-					free(module);
-					continue;
-				}
-				start = ptr;
-				while (*ptr && *ptr != ' ') ptr++;
-				if (*ptr) {
-					*ptr = '\0';
-					ptr++;
-				}
-				subdevid = strtoul(start, (char **)NULL, 16);
-				if (subdevid == 0xffffffff)
-				  subdevid = 0;
-			}
+		if (*start == '#') {
 			start = buf;
-
-			tmpdev = nextDevice;
+			continue;
+		}
+		ptr = start;
+		while (*ptr && *ptr != ' ') ptr++;
+		if (*ptr) {
+			*ptr = '\0';
+			ptr++;
+		}
+		module = strdup(start);
+		while (*ptr && *ptr == ' ') ptr++;
+		if (!*ptr) {
+			start = buf;
+			free(module);
+			continue;
+		}
+		start = ptr;
+		while (*ptr && *ptr != ' ') ptr++;
+		if (*ptr) {
+			*ptr = '\0';
+			ptr++;
+		}
+		vendid = strtoul(start,(char **)NULL,16);
+		if (vendid == 0xffffffff) {
+			start = buf;
+			free(module);
+			continue;
+		}
+		while (*ptr && *ptr == ' ') ptr++;
+		if (!*ptr) {
+			start = buf;
+			free(module);
+			continue;
+		}
+		start = ptr;
+		while (*ptr && *ptr != ' ') ptr++;
+		if (*ptr) {
+			*ptr = '\0';
+			ptr++;
+		}
+		devid = strtoul(start, (char **)NULL, 16);
+		while (*ptr && *ptr == ' ') ptr++;
+		if (!*ptr) {
+			start = buf;
+			free(module);
+			continue;
+		}
+		start = ptr;
+		while (*ptr && *ptr != ' ') ptr++;
+		if (*ptr) {
+			*ptr = '\0';
+			ptr++;
+		}
+		subvendid = strtoul(start, (char **)NULL, 16);
+		if (subvendid == 0xffffffff) {
+			subvendid = 0xffff;
+			subdevid = 0;
+		} else {
+			while (*ptr && *ptr == ' ') ptr++;
+			if (!*ptr) {
+				start = buf;
+				free(module);
+				continue;
+			}
+			start = ptr;
+			while (*ptr && *ptr != ' ') ptr++;
+			if (*ptr) {
+				*ptr = '\0';
+				ptr++;
+			}
+			subdevid = strtoul(start, (char **)NULL, 16);
+			if (subdevid == 0xffffffff)
+				subdevid = 0;
+		}
+		start = buf;
+		
+		tmpdev = nextDevice;
+		key.vendorId = vendid;
+		key.deviceId = devid;
+		key.subVendorId = subvendid;
+		key.subDeviceId = subdevid;
+		key.pciType = PCI_NORMAL;
+		nextDevice = bsearch(&key,pciDeviceList,numPciDevices,
+				     sizeof(struct pciDevice), devCmp);
+		/* HACK */
+		if (nextDevice && !strcmp(nextDevice->driver, "bcm5700"))
+			nextDevice = NULL;
+		/* HACK */
+		if (nextDevice && !strcmp(nextDevice->driver, "eepro100"))
+			nextDevice = NULL;
+		if (!nextDevice) {
 			key.vendorId = vendid;
 			key.deviceId = devid;
-			key.subVendorId = subvendid;
-			key.subDeviceId = subdevid;
+			key.subVendorId = 0xffff;
+			key.subDeviceId = 0;
 			key.pciType = PCI_NORMAL;
-			nextDevice = bsearch(&key,pciDeviceList,numPciDevices,
-					     sizeof(struct pciDevice), devCmp);
-			if (!nextDevice) {
-				key.vendorId = vendid;
-				key.deviceId = devid;
-				key.subVendorId = 0xffff;
-				key.subDeviceId = 0;
-				key.pciType = PCI_NORMAL;
-				nextDevice = bsearch(&key, pciDeviceList, numPciDevices,
-						     sizeof (struct pciDevice), devCmp);
-				if (nextDevice && strcmp(nextDevice->driver, module))
-			     		nextDevice = NULL;
-			}
-		    	if (!nextDevice) {
-				char newdesc[64];
-				nextDevice = tmpdev;
-				tmpdev = NULL;
-				numPciDevices++;
-				nextDevice->vendorId = vendid;
-				nextDevice->deviceId = devid;
-				nextDevice->subVendorId = subvendid;
-				nextDevice->subDeviceId = subdevid;
-				nextDevice->pciType = PCI_NORMAL;
-				nextDevice->driver = strdup(module);
-				snprintf(newdesc,63,"Unknown vendor|Generic %s device",module);
-				nextDevice->desc = strdup(newdesc);
-				nextDevice->next = NULL;
-				nextDevice->device = NULL;
-				nextDevice->class = 0;
-				nextDevice->bus = BUS_PCI;
-			} else {
-				if (!strcmp(nextDevice->driver,"unknown") ||
-				    !strcmp(nextDevice->driver,"disabled") ||
-				    !strcmp(nextDevice->driver,"ignore")) {
-					free(nextDevice->driver);
-					nextDevice->driver = strdup(module);
-				}
-			}
-			if (tmpdev)
-			  nextDevice = tmpdev;
-			else {
-				nextDevice++;
-				qsort(pciDeviceList, numPciDevices, sizeof(*pciDeviceList), devCmp);
-			}
-			free(module);
+			nextDevice = bsearch(&key, pciDeviceList, numPciDevices,
+					     sizeof (struct pciDevice), devCmp);
 		}
-		free(bufptr);
+		if (!nextDevice) {
+			char newdesc[64];
+			nextDevice = tmpdev;
+			tmpdev = NULL;
+			numPciDevices++;
+			nextDevice->vendorId = vendid;
+			nextDevice->deviceId = devid;
+			nextDevice->subVendorId = subvendid;
+			nextDevice->subDeviceId = subdevid;
+			nextDevice->pciType = PCI_NORMAL;
+			nextDevice->driver = strdup(module);
+			snprintf(newdesc,63,"Unknown vendor|Generic %s device",module);
+			nextDevice->desc = strdup(newdesc);
+			nextDevice->next = NULL;
+			nextDevice->device = NULL;
+			nextDevice->type = 0;
+			nextDevice->bus = BUS_PCI;
+		} else {
+			if (!strcmp(nextDevice->driver,"unknown") ||
+			    !strcmp(nextDevice->driver,"disabled") ||
+			    !strcmp(nextDevice->driver,"ignore")) {
+				free(nextDevice->driver);
+				nextDevice->driver = strdup(module);
+			}
+		}
+		if (tmpdev)
+			nextDevice = tmpdev;
+		else {
+			nextDevice++;
+			qsort(pciDeviceList, numPciDevices, sizeof(*pciDeviceList), devCmp);
+		}
+		free(module);
 	}
+	free(bufptr);
 	return 0;
 }
 
@@ -658,19 +690,10 @@ static int isDisabled(struct pci_dev *p, byte config[256]) {
 	unsigned int devtype, command;
 	
 	devtype = config[PCI_CLASS_DEVICE+1] << 8 | config[PCI_CLASS_DEVICE];
-	/* Network, USB, and audio devices with no IRQ are
-	 * generally 'disabled-in-bios' */
-	if ( (p->irq == 0 || p->irq == 255) &&
-	    (pciToKudzu(devtype) == CLASS_NETWORK ||
-	     pciToKudzu(devtype) == CLASS_USB ||
-	     pciToKudzu(devtype) == CLASS_AUDIO
-	     )) {
-		return 1;
-	}
-	if (p->irq) {
+	if (p->irq || pciToKudzu(devtype) != CLASS_VIDEO) {
 		return 0;
 	}
-	/* Ditto for things that are explicitly disabled */
+	/* Only check video cards for now. */
 	command = config[PCI_COMMAND];
 	disabled = 0;
 	for (i = 0; i < limit ; i++) {
@@ -697,17 +720,19 @@ struct device * pciProbe(enum deviceClass probeClass, int probeFlags, struct dev
     int init_list = 0;
     unsigned int type = kudzuToPci(probeClass),devtype;
     
-    if ((probeClass==CLASS_UNSPEC) ||
-	(probeClass==CLASS_OTHER) ||
-	(probeClass==CLASS_NETWORK) ||
-	(probeClass==CLASS_SCSI) ||
-	(probeClass==CLASS_VIDEO) ||
-	(probeClass==CLASS_AUDIO) ||
-	(probeClass==CLASS_MODEM) ||
-	(probeClass==CLASS_USB) ||
-	(probeClass==CLASS_SOCKET) ||
-	(probeClass==CLASS_CAPTURE) ||
-	(probeClass==CLASS_RAID)) {
+    if (
+	(probeClass & CLASS_OTHER) ||
+	(probeClass & CLASS_NETWORK) ||
+	(probeClass & CLASS_SCSI) ||
+	(probeClass & CLASS_IDE) ||
+	(probeClass & CLASS_VIDEO) ||
+	(probeClass & CLASS_AUDIO) ||
+	(probeClass & CLASS_MODEM) ||
+	(probeClass & CLASS_USB) ||
+	(probeClass & CLASS_FIREWIRE) ||
+	(probeClass & CLASS_SOCKET) ||
+	(probeClass & CLASS_CAPTURE) ||
+	(probeClass & CLASS_RAID)) {
 	pacc = pci_alloc();
 	if (!pacc) return devlist;
 	if (!pciDeviceList) {
@@ -754,6 +779,7 @@ struct device * pciProbe(enum deviceClass probeClass, int probeFlags, struct dev
 		    subvend = config[PCI_SUBSYSTEM_VENDOR_ID+1] << 8 | config[PCI_SUBSYSTEM_VENDOR_ID];
 		    subdev = config[PCI_SUBSYSTEM_ID+1] << 8 | config[PCI_SUBSYSTEM_ID];
 		}
+		pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
 		bustype = PCI_NORMAL;
 	        for (bridgenum=0; cardbus_bridges[bridgenum]; bridgenum++) {
 			if (p->bus == cardbus_bridges[bridgenum])
@@ -765,13 +791,26 @@ struct device * pciProbe(enum deviceClass probeClass, int probeFlags, struct dev
 			/* Test to see if it's UHCI or OHCI */
 			free (dev->driver);
 			if (config[PCI_CLASS_PROG] == 0) {
-				dev->driver = strdup("usb-uhci");
+				if (kernel_release >= 2.5) 
+					dev->driver = strdup("uhci-hcd");
+				else
+					dev->driver = strdup("usb-uhci");
 			} else if (config[PCI_CLASS_PROG] == 0x10) {
-				dev->driver = strdup("usb-ohci");
+				if (kernel_release >= 2.5)
+					dev->driver = strdup("ohci-hcd");
+				else
+					dev->driver = strdup("usb-ohci");
 			} else if (config[PCI_CLASS_PROG] == 0x20) {
 				dev->driver = strdup("ehci-hcd");
 			}
 		}
+		if (devtype == PCI_CLASS_SERIAL_FIREWIRE) {    
+			/* Test to see if it's OHCI */
+			if (config[PCI_CLASS_PROG] == 0x10) {
+				free (dev->driver);
+				dev->driver = strdup("ohci1394");
+			}
+		}
 		/* Check for an i2o device. Note that symbios controllers
 		 * also need i2o_scsi module. Dunno how to delineate that
 		 * here. */
@@ -793,13 +832,23 @@ struct device * pciProbe(enum deviceClass probeClass, int probeFlags, struct dev
 	        if (dev->pciType == PCI_CARDBUS) {
 			dev->detached = 1;
 		}
+		/* Sure, reuse a pci id for an incompatible card. */
+		if (dev->vendorId == 0x10ec && dev->deviceId == 0x8139 &&
+		    config[PCI_REVISION_ID] >= 0x20) {
+			free(dev->driver);
+			dev->driver = strdup("8139cp");
+		}
+		dev->pcidom = p->domain;
+		dev->pcibus = p->bus;
+		dev->pcidev = p->dev;
+		dev->pcifn = p->func;
 			
 		if ( (probeFlags & PROBE_ALL) || (strcmp(dev->driver,"unknown") && strcmp(dev->driver, "disabled") && strcmp(dev->driver,"ignore"))) {
 		    if (!type || (type<0xff && (type==devtype>>8))
 			|| (type== kudzuToPci (pciToKudzu (devtype)))) {
 			a_dev = pciNewDevice(dev);
-			a_dev->class = pciToKudzu(devtype);
-			switch (a_dev->class) {
+			a_dev->type = pciToKudzu(devtype);
+			switch (a_dev->type) {
 			case CLASS_NETWORK:
 				if (devtype == PCI_CLASS_NETWORK_TOKEN_RING)
 				  a_dev->device = strdup("tr");
