@@ -1,3 +1,14 @@
+/* Copyright 1999-2003 Red Hat, Inc.
+ *
+ * This software may be freely redistributed under the terms of the GNU
+ * public license.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ */
+
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
@@ -8,6 +19,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <linux/cdrom.h>
+#include <sys/ioctl.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
 #include <sys/types.h>
@@ -24,8 +37,10 @@
 
 #define PROGRAM_NAME	"updfstab"
 
+#define CLASS_RESERVED_HACK 0
+
 struct matchInfo {
-    int class;
+    int type;
     const char * descPattern;
 };
 
@@ -34,6 +49,8 @@ struct typeInfo {
     int makeSymlink;
     int autoMount;
     int skip;
+    int noFstab;
+    int ignoreMulti;
     int defaultPartition;
     struct matchInfo * matches;
     int numMatches;
@@ -67,7 +84,10 @@ struct partitionEntry {
     int partition;
 };
 
+static struct stat fstabsb;
+
 static int readMountTable(char * fileName, struct mountInfo ** mountListPtr);
+static int isWritable(char * path);
 static int getDeviceNum(char * path, dev_t * device);
 static int probeDeviceType(struct partitionEntry * partList,
 			   const struct typeInfo * type, 
@@ -77,6 +97,7 @@ static int removeDeadDevices(struct mountInfo * mountTable,
 			     struct probedDevice * deviceList);
 static int probeAllDevices(struct partitionEntry * partList,
 			   struct typeInfo * deviceInfoList,
+			   int skipprobe,
 			   struct probedDevice ** list);
 static int buildAddList(struct mountInfo * mountedDevices, 
 			struct mountInfo * mountTable, 
@@ -129,6 +150,27 @@ static dev_t fullDiskDevice(dev_t dev) {
     return makedev(maj, (min & ~mask));
 }
 
+static int isWritable(char *dev) {
+    int fd, capability;
+    char path[50];
+    
+    snprintf(path,50,"/dev/%s",dev);
+    
+    fd = open(path, O_RDONLY|O_NONBLOCK);
+    if (fd == -1) return 0;
+    capability = ioctl (fd, CDROM_GET_CAPABILITY, &capability);
+    close(fd);
+    if (capability == -1) return 0;
+    if (capability & (
+		      0x2000 |  /* CD-R */
+		      0x4000 |  /* CD-RW */
+		      0x10000 | /* DVD-R */
+		      0x20000   /* DVD-RAM */
+		      ))
+        return 1;
+    return 0;
+}
+
 static int getDeviceNum(char * path, dev_t * device) {
     struct stat sb;
 
@@ -213,13 +255,27 @@ static int probeDeviceType(struct partitionEntry * partList,
 
 	if (!type->matches)
 	    keepDevice = 0;
+	    
+        if (!(aDevice->bus == BUS_SCSI ||
+	      aDevice->bus == BUS_IDE ||
+	      aDevice->bus == BUS_MISC ||
+	      aDevice->bus == BUS_FIREWIRE))
+	    keepDevice = 0;
+	
+	if (!aDevice->device)
+	    keepDevice = 0;
 
 	if (keepDevice) {
 	    int i;
 
 	    keepDevice = 0;
 	    for (i = 0; i < type->numMatches; i++) {
-		if (aDevice->class == type->matches[i].class) {
+		if (aDevice->type == type->matches[i].type ||
+		    /* CD-RW hack! */
+		    ( aDevice->type == CLASS_CDROM &&
+		      isWritable(aDevice->device) &&
+		      type->matches[i].type == CLASS_RESERVED_HACK )
+		    ) {
 		    if (!type->matches[i].descPattern ||
 		      strcasestr(aDevice->desc, type->matches[i].descPattern)) {
 			keepDevice = 1;
@@ -236,13 +292,13 @@ static int probeDeviceType(struct partitionEntry * partList,
 	    /* if there are multiple partitions on a device, bail. note
 	       the kernel sorts /proc/partitions for us */
 	    if (part->device && part[1].device &&
-		!strcmp(part[1].device, aDevice->device)) {
+		!strcmp(part[1].device, aDevice->device) && !type->ignoreMulti) {
 		continue;
 	    }
-
+		
 	    devName = malloc(strlen(aDevice->device) + 10);
 	    
-	    if (part->device)
+	    if (part->device && !type->ignoreMulti)
 		sprintf(devName, "/dev/%s%d", part->device,
 			part->partition);
 	    else if (type->defaultPartition)
@@ -252,6 +308,7 @@ static int probeDeviceType(struct partitionEntry * partList,
 		sprintf(devName, "/dev/%s", aDevice->device);
 	    
 	    /* possibly already probed this device w/a different pattern */
+	    if (type->matches[0].type != CLASS_RESERVED_HACK)
 	    for (flag = 0, devListPtr = *list; devListPtr;
 		 devListPtr = devListPtr->next) {
 		if (!strcmp(devListPtr->device, devName)) {
@@ -270,7 +327,8 @@ static int probeDeviceType(struct partitionEntry * partList,
 	    newDevice->type = type;
 	    newDevice->next = NULL;
 
-	    if (getDeviceNum(newDevice->device, &newDevice->devNum))
+	    if (!newDevice->type->noFstab &&
+		getDeviceNum(newDevice->device, &newDevice->devNum))
 		return 1;
 
 	    if (endOfList) {
@@ -288,12 +346,21 @@ static int probeDeviceType(struct partitionEntry * partList,
 
 static int probeAllDevices(struct partitionEntry * partList,
 			   struct typeInfo * deviceInfoList,
+			   int skipprobe, 
 			   struct probedDevice ** list) {
     const struct typeInfo * aType;
     struct device ** kudzuDevices;
     int x;
+    struct stat sbuf;
 
-    kudzuDevices = probeDevices(CLASS_UNSPEC, BUS_IDE | BUS_SCSI | BUS_MISC, 0);
+    if (skipprobe) {
+	if (stat("/etc/sysconfig/hwconf", &sbuf) == -1) return 0;
+	kudzuDevices = readDevices("/etc/sysconfig/hwconf");
+    } else {
+	kudzuDevices = probeDevices(CLASS_CDROM | CLASS_HD | CLASS_FLOPPY,
+				    BUS_IDE | BUS_SCSI |
+				    BUS_MISC | BUS_FIREWIRE, 0);
+    }
     if (!kudzuDevices) return 0;
     for (aType = deviceInfoList; aType->name; aType++)
 	if (probeDeviceType(partList, aType, kudzuDevices, list)) return 1;
@@ -513,36 +580,17 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 	return 1;
     }
 
-    if (test) {
-	output = stdout;
-	outputFileName = NULL;
-
-	revoke = stdout;
-	revokeFileName = NULL;
-    } else {
-	outputFileName = alloca(strlen(fileName) + 5);
-	strcpy(outputFileName, fileName);
-	strcat(outputFileName, ".NEW");
-
-	if (!(output = fopen(outputFileName, "w"))) {
-	    fprintf(stderr, _("failed to open %s: %s\n"),
-		    outputFileName, strerror(errno));
-	    return 1;
-	}
-
-	revokeFileName = alloca(strlen(fileName) + 10);
-	strcpy(revokeFileName, fileName);
-	strcat(revokeFileName, ".REVOKE");
-
-	if (!(revoke = fopen(revokeFileName, "w"))) {
-	    fprintf(stderr, _("failed to open %s: %s\n"),
-		    revokeFileName, strerror(errno));
-	    return 1;
-	}
-    }
-
     fstat(fd, &before);
 
+    if (before.st_mtime != fstabsb.st_mtime ||
+	    before.st_dev != fstabsb.st_dev ||
+	    before.st_ino != fstabsb.st_ino ||
+	    before.st_size != fstabsb.st_size) {
+	    fprintf(stderr, _("%s changed during run -- aborting\n"),
+		    fileName);
+	    return 1;
+    }
+
     fstab = alloca(before.st_size + 2);
     if (read(fd, fstab, before.st_size) != before.st_size) {
 	fprintf(stderr, _("failed to read %s: %s\n"), fileName, 
@@ -560,6 +608,37 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 	fstab[before.st_size + 1] = '\0';
     }
 
+    if (test) {
+	output = stdout;
+	outputFileName = NULL;
+
+	revoke = stdout;
+	revokeFileName = NULL;
+    } else {
+	outputFileName = alloca(strlen(fileName) + 12);
+	strcpy(outputFileName, fileName);
+	strcat(outputFileName, ".NEW.XXXXXX");
+
+	if ((fd = mkstemp(outputFileName)) < 0 ||
+	    !(output = fdopen(fd, "w"))) {
+	    fprintf(stderr, _("failed to open %s: %s\n"),
+		    outputFileName, strerror(errno));
+	    return 1;
+	}
+	chmod(outputFileName,0644);
+
+	revokeFileName = alloca(strlen(fileName) + 17);
+	strcpy(revokeFileName, fileName);
+	strcat(revokeFileName, ".REVOKE.XXXXXX");
+
+	if ((fd = mkstemp(revokeFileName)) < 0 ||
+	    !(revoke = fdopen(fd, "w"))) {
+	    fprintf(stderr, _("failed to open %s: %s\n"),
+		    revokeFileName, strerror(errno));
+	    return 1;
+	}
+    }
+
     lineStart = fstab;
 
     /* copy what's there, removing things which need to disappear */
@@ -589,6 +668,7 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 
 		if (getDeviceNum(dev, &devNum)) {
 		    if (outputFileName) unlink(outputFileName);
+		    if (revokeFileName) unlink(revokeFileName);
 		    return 1;
 		}
 
@@ -641,6 +721,7 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 			    symlinkSource, addition->device->device, 
 			    strerror(errno));
 		    if (outputFileName) unlink(outputFileName);
+		    if (revokeFileName) unlink(revokeFileName);
 		    return 1;
 		}
 	    }
@@ -649,27 +730,31 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 	}
 
 	if (!test) {
-	    if (mkdir(addition->mountPoint, 0775) && errno != EEXIST) {
+	    if (!addition->device->type->noFstab &&
+		mkdir(addition->mountPoint, 0775) && errno != EEXIST) {
 		fprintf(stderr, _("failed to create %s: %s\n"),
 			addition->mountPoint, strerror(errno));
 
 		if (outputFileName) unlink(outputFileName);
+		if (revokeFileName) unlink(revokeFileName);
 		return 1;
 	    }
 	}
 
-	fprintf(output, "%-23s %-23s %-7s %-15s %d %d\n",
-		deviceName,
-		addition->mountPoint,
-		!(strcmp(addition->device->type->name, "cdrom") &&
-		  strcmp(addition->device->type->name, "cdwriter") &&
-		  strcmp(addition->device->type->name, "cdrw")) ?
-		"iso9660" : "auto",
-		!(strcmp(addition->device->type->name, "cdrom") &&
-		  strcmp(addition->device->type->name, "cdwriter") &&
-		  strcmp(addition->device->type->name, "cdrw")) ?
-		"noauto,owner,kudzu,ro" : "noauto,owner,kudzu",
-		0, 0);
+	/* CD-RW hack! */
+	if (!addition->device->type->noFstab)
+	    fprintf(output, "%-23s %-23s %-7s %-15s %d %d\n",
+		    deviceName,
+		    addition->mountPoint,
+		    !(strcmp(addition->device->type->name, "cdrom") &&
+		      strcmp(addition->device->type->name, "cdwriter") &&
+		      strcmp(addition->device->type->name, "cdrw")) ?
+		    "udf,iso9660" : "auto",
+		    !(strcmp(addition->device->type->name, "cdrom") &&
+		      strcmp(addition->device->type->name, "cdwriter") &&
+		      strcmp(addition->device->type->name, "cdrw")) ?
+		    "noauto,owner,kudzu,ro" : "noauto,owner,kudzu",
+		    0, 0);
     }
 
     if (!test) {
@@ -684,6 +769,7 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 	    fprintf(stderr, _("%s changed during run -- aborting\n"),
 		    fileName);
 	    unlink(outputFileName);
+	    unlink(revokeFileName);
 
 	    return 1;
 	}
@@ -718,10 +804,12 @@ static int updateFstab(char * fileName, struct mountInfo * mountTable,
 	if (rename(outputFileName, fileName)) {
 	    fprintf(stderr, _("failed to rename %s to %s: %s\n"),
 		    outputFileName, fileName, strerror(errno));
+	    unlink(revokeFileName);
 	    unlink(outputFileName);
 	    return 1;
 	}
     }
+    if (revokeFileName) unlink(revokeFileName);
 
     return 0;
 }
@@ -739,18 +827,29 @@ static struct partitionEntry * readPartitionList(void) {
     char lastFullDevice[256];
     char *iobuf = NULL;
     long long lastBlockCount = -1;
-
-    f = fopen("/proc/partitions", "r");
-    if (!f) {
+    int len, fd;
+    
+    fd = open("/proc/partitions", O_RDONLY, 0);
+    if (fd < 0) {
 	fprintf(stderr, _("failed to open /proc/partitions: %s\n"),
 		strerror(errno));
 	return NULL;
     }
     iobuf = (char *)malloc(16384);
-    if (iobuf) {
-	    setvbuf(f, iobuf, _IOFBF, 16384);
+    if (!iobuf) {
+	    fprintf(stderr, _("out of memory.\n"));
+	    return NULL;
     }
-
+    len = read(fd, iobuf, 16384);
+    if (len < 0) {
+	fprintf(stderr, _("failed to read /proc/partitions: %s\n"),
+		strerror(errno));
+	close(fd);
+	return NULL;
+    }
+	    
+    f = fmemopen(iobuf, len, "r");
+    
     /* first two lines are junk */
     frc = fgets(line, sizeof(line), f);
     if (frc) frc = fgets(line, sizeof(line), f);
@@ -805,7 +904,8 @@ static struct partitionEntry * readPartitionList(void) {
     list[numItems++].device = NULL;
 
     fclose(f);
-    if (iobuf) free(iobuf);
+    free(iobuf);
+    close(fd);
 
     return list;
 }
@@ -823,8 +923,10 @@ struct cfInfo {
 struct cfInfo cfOptions[] = {
     /* {	"automount",	0,  BOOLEAN,    NONE },*/
     {	"device",	0,  STRING,	STRING },
+    {	"ignoremulti",  0,  BOOLEAN,    NONE },
     {	"include",	0,  STRING,	NONE  },
     {	"match",	1,  STRING,	STRINGOPT },
+    {	"nofstab",	0,  BOOLEAN,	NONE  },
     {	"partition",	0,  INTEGER,    NONE  },
     {	"skip",		0,  BOOLEAN,    NONE  },
     {	"symlink",	0,  BOOLEAN,    NONE  },
@@ -847,7 +949,7 @@ struct typeInfo * readConfigFile(const char * path,
     const char ** lineArgv;
     int i, j;
     int intArg1;
-    struct typeInfo defaultStruct = { NULL, 0, 0, 0, 0, NULL, 0 };
+    struct typeInfo defaultStruct = { NULL, 0, 0, 0, 0, 0, 0, NULL, 0 };
     struct typeInfo * item;
 
     if (!defaults) 
@@ -980,8 +1082,12 @@ struct typeInfo * readConfigFile(const char * path,
 	    item->autoMount = intArg1;
 	} else if (!strcmp(lineArgv[0], "skip")) {
 	    item->skip = intArg1;
+	} else if (!strcmp(lineArgv[0], "nofstab")) {
+	    item->noFstab = intArg1;	
 	} else if (!strcmp(lineArgv[0], "symlink")) {
 	    item->makeSymlink = intArg1;
+	} else if (!strcmp(lineArgv[0], "ignoremulti")) {
+            item->ignoreMulti = intArg1;
 	} else if (!strcmp(lineArgv[0], "device")) {
 	    if (strcmp(lineArgv[2], "{")) {
 		fprintf(stderr, "%s:%d { expected\n", path, line);
@@ -1020,11 +1126,14 @@ struct typeInfo * readConfigFile(const char * path,
 		item->matches[i].descPattern = NULL;
 
 	    if (!strcmp(lineArgv[1], "hd"))
-		item->matches[i].class = CLASS_HD;
+		item->matches[i].type = CLASS_HD;
 	    else if (!strcmp(lineArgv[1], "cdrom"))
-		item->matches[i].class = CLASS_CDROM;
+		item->matches[i].type = CLASS_CDROM;
 	    else if (!strcmp(lineArgv[1], "floppy"))
-		item->matches[i].class = CLASS_FLOPPY;
+		item->matches[i].type = CLASS_FLOPPY;
+	    /* CD-RW hack! */
+	    else if (!strcmp(lineArgv[1], "cdwriter"))
+	        item->matches[i].type = CLASS_RESERVED_HACK;
 	    else {
 		fprintf(stderr, "%s:%d unknown device class %s\n",
 			path, line, lineArgv[1]);
@@ -1056,7 +1165,7 @@ int main(int argc, const char * argv[]) {
     struct entryToAdd * addList = NULL;
     struct partitionEntry * partitionList;
     struct typeInfo * types = NULL;
-    int test = 0, normalize = 0;
+    int test = 0, normalize = 0, skipprobe = 0;
     int rc;
     poptContext optCon;
     char * configFile = FSTAB_DEVICES;
@@ -1066,7 +1175,9 @@ int main(int argc, const char * argv[]) {
 		"path" },
 	{ "normalize", 'n', 0, &normalize, 0, 
 		"update fstab entries to match device probing order" },
-	{ "test", 't', 0, &test, 0,
+	{ "skipprobe", 's', 0, &skipprobe, 0,
+	  "skip probing for new devices and read stored configuration instead" },
+	{ "test", 't',0, &test, 0,
 		"don't update anything, display new fstab on stdout" },
 	POPT_AUTOHELP
 	{ NULL, '\0', 0, NULL },
@@ -1084,14 +1195,20 @@ int main(int argc, const char * argv[]) {
 
     poptFreeContext(optCon);
 
+    if (access(FSTAB, W_OK) && !test) {
+	perror("Unable to open /etc/fstab for writing");
+	return 1;
+    }
+
     types = calloc(sizeof(*types), 1);
     if (!(types = readConfigFile(configFile, types, NULL))) return 1;
 
     if (readMountTable(MTAB, &mountList)) return 1;
+    stat(FSTAB,&fstabsb);
     if (readMountTable(FSTAB, &mountTable)) return 1;
 
     if (!(partitionList = readPartitionList())) return 1;
-    if (probeAllDevices(partitionList, types, &devices)) return 1;
+    if (probeAllDevices(partitionList, types, skipprobe, &devices)) return 1;
 
     if (removeDeadDevices(mountTable, devices)) return 1;
     if (buildAddList(mountList, mountTable, devices, types, &addList, 
